(** Make a pure data structure module buildable from imperative events. *)

(** Input signature for the module to be used from imperative changes.
    It implements the data structure that will be usable through FRP.  *)
module type M = sig
  type t
  (** The type of the data structure. *)
  type init
  (** The type of the argument passed to the M.create function. *)
  type action_functional
  (** Actions indicate how to change the data structure: for instance, "`Add 42"
      to add the int 42 to a set data structure.
      action_functional is for actions to do in the functional side of the
      program (triggered from the UI). *)
  type action_ui
  (** See action_functional. action_ui is for actions to do in the UI (triggered
      from the functional side of the program). *)
  val create : init -> t
  (** The initial state the data structure will be in. *)
  val on_action_func : t -> action_functional -> (t * action_ui) Lwt.t
  (** on_action is a function passed to a "fold" operation: from the current
   state and an action, it creates a new state.
   Here, 'a is an identifier for the source of the action. It is opaque but can
   be used to find whether the source of the action was the same as the source
   of a previous action (which is why it is stored in the datastructure). *)
end

(** Makes a pure data structure usable from imperative events. *)
module Core : functor (M : M) -> sig
  type callback = M.action_ui -> unit
  (** Type of functions to be called each time the data structure has been
   updated. *)
  type propagator = M.t -> unit
  (** Type of functions to be called each time the data structure has been
   updated and the data structure containing it (if any) (nested data
   structures) has to be updated. *)
  class core : ?propagators:(propagator list) -> M.init -> object
    method connect : callback -> unit
    (** Add a callback to be called when the functional state changes. *)
    method send : M.action_functional -> unit
    (** Send an action in order to make the functional state change. *)
    method propagator_add : propagator -> unit
    (** Add a propagator. *)
    method propagator_rm : propagator -> unit
    (** Remove a propagator. *)
  end
end

