(* n_rev_append l1 l2 n :
  * move n elements from l1 to l2, reversing their order. *)
let rec n_rev_append l1 l2 n =
  if n = 0 then
    l1, l2
  else
    match l1 with
    | [] -> l1, l2
    | h :: t -> n_rev_append t (h :: l2) (n - 1)

type 'a t = { l : 'a list; r : 'a list; ll : int; lr : int }

type where_add = [ `Beginning | `End | `Before | `After | `Here | `At of int ]
type where_remove = [ `Before | `After ]

exception End_of_zipper

let empty = { l = []; r = []; ll = 0; lr = 0 }

let next { l; r; ll; lr } =
  match r with
  | [] | [ _ ] -> raise End_of_zipper
  | h :: q -> { l = h :: l; r = q; ll = ll + 1; lr = lr - 1 }

let previous { l; r; ll; lr } =
  match l with
  | h :: q -> { l = q; r = h :: r; ll = ll - 1; lr = lr + 1 }
  | [] -> raise End_of_zipper

let first { l; r; ll; lr } =
  let right = List.rev_append l r in
  { l = []; r = right; ll = 0; lr = ll + lr }

let last { l; r; ll; lr } =
  let left = List.rev_append r l in
  { l = left; r = []; ll = 0; lr = ll + lr }

let pos { ll } =
  ll

let length { ll; lr } =
  ll + lr

let get = function
  | { r = h :: _ } -> h
  | _ -> raise End_of_zipper

let set z e =
  match z.r with
  | h :: tl -> { z with r = e :: tl }
  | _ -> raise End_of_zipper

let nth ({ l; r; ll; lr } as z) i =
  if i < 0 || i >= length z then
    raise (Invalid_argument "index out of bounds")
  else
    if i = pos z then
      z
    else
      if i > ll then
        let r, l = n_rev_append r l (i - ll) in
        { l; r; ll = i; lr = lr - i + ll }
      else
        let l, r = n_rev_append l r (ll - i) in
        { l; r; ll = i; lr = lr + ll - i }

let add_before ({ l; ll } as z) e =
  { z with l = e :: l; ll = ll + 1 }

let add_here ({ r; lr } as z) e =
  { z with r = e :: r; lr = lr + 1 }

let add_after ({ r; lr } as z) e =
  match r with
  | [] -> { z with r = [ e ]; lr = 1 }
  | h :: q -> { z with r = h :: e :: q; lr = lr + 1 }

let add_at z i e =
  let pos = pos z in
  if i = pos then
    add_here z e
  else
    if i < pos then
      let z = nth z i in
      let z = add_before z e in
      nth z (pos + 1)
    else
      let z = nth z (i - 1) in (* i > pos >= 0 which implies i > 0 *)
      let z = add_after z e in
      nth z pos

let add (where : where_add) z e =
  match where with    
  | `Beginning -> add_at z 0 e
  | `End -> add_at z (length z) e
  | `Before -> add_before z e
  | `After -> add_after z e
  | `Here -> add_here z e
  | `At i -> add_at z i e

let remove (new_focus : where_remove) ({ l; r; lr; ll } as z) =
  match l, r with
  | _, [] -> raise End_of_zipper
  | [], r_hd :: r_tl -> { z with r = r_tl; lr = lr - 1 }
  | l_hd :: l_tl, [ _ ] -> { r = [ l_hd ]; lr = 1; l = l_tl; ll = ll - 1 }
  | l_hd :: l_tl, r_hd :: r_tl -> begin
      match new_focus with
      | `After -> { z with r = r_tl; lr = lr - 1 }
      | `Before -> { r = l_hd :: r_tl; lr = lr; l = l_tl; ll = ll - 1 }
    end

let remove_nth set i =
  let pos = pos set in
  let target = if i < pos then pos - 1 else pos in
  nth (remove `After (nth set i)) target

let rec move z i j =
  let zi = nth z i in
  let i_data = get zi in
  let z_without_i = remove `After zi in
  if i < j then
    let z = nth z_without_i (j - 1) in
    add_after z i_data
  else
    let z = nth z_without_i j in
    add_before z i_data

(* find predicate zipper:
  * Let zipper be { l = [ 5; 4; 3; 2; 1 ]; r = [ 6; 7 ] }
  * The main idea is to look alternatively to the left and to the right
  * (sub-function "find_bidir"). The
  * order is as follows:
    * 5, 6, 4, 7, 3
  * Most of the time, alternating will become impossible after some time. The
  * program will then stop alternating and go straight in one direction
  * (sub-function "find_unidir") *)
let find pred set =
  let rec find_unidir move set =
    if pred set then
      set
    else
      find_unidir move (move set)
  in
  let rec find_bidir (move1, set1, move2, set2) =
    if pred set1 then
      set1
    else
      (* If we can't go further on set1, we've reach the end of this direction
       * and we switch to find_unidir. *)
      if pos set1 = 0 || pos set1 = length set1 then
        find_unidir move2 set2
      else
        find_bidir (move2, set2, move1, (move1 set1))
  in
  let pos = pos set in
  try 
    if pos = 0 then
      find_unidir next set
    else
      find_bidir (previous, (previous set), next, set)
  with End_of_zipper -> raise Not_found

